from textual import work
from textual.app import App, ComposeResult

from textual.widgets import Footer, Header

from blueutil_tui.utils import get_blueutil_version
from blueutil_tui.widgets import OverViewTable


class BlueUtilApp(App):
    def on_mount(self):
        self.get_blueutil_version()

    def compose(self) -> ComposeResult:
        self.screen.title = "blueutil-tui"
        yield Header(icon="")
        yield OverViewTable()
        yield Footer()
        return super().compose()

    @work(thread=True)
    def get_blueutil_version(self):
        version = get_blueutil_version()
        self.screen.title += f" using blueutil v{version}"



from blueutil_tui.app import BlueUtilApp
from blueutil_tui.utils import (
    check_blueutil_installation,
    # search_new_devices,
    # get_paired_devices,
)


def main() -> None:
    if not check_blueutil_installation():
        return
    # devices = search_new_devices()
    # for dev in devices:
    #     print(dev)

    app = BlueUtilApp()
    app.run(inline=True)



import subprocess
import shutil
import json

from rich.console import Console

console = Console()
TIMEOUT = 8


def check_blueutil_installation():
    if shutil.which("blueutil") is None:
        console.print(
            '[blue]"blueutil"[/] was not found, please install with e.g. [blue]"brew install blueutil"[/] '
        )
        console.print("or use another installation method from:")
        console.print(
            "[blue underline]https://github.com/toy/blueutil?tab=readme-ov-file#installupdateuninstall[/] "
        )
        return False
    return True


def get_blueutil_version():
    command = subprocess.run(
        args=["blueutil", "--version"],
        capture_output=True,
        text=True,
    )
    return command.stdout.strip()


def get_paired_devices():
    command = subprocess.run(
        args=["blueutil", "--paired", "--format", "json"],
        capture_output=True,
        text=True,
        timeout=TIMEOUT,
    )

    handle_returncodes(errorcode=command.returncode)

    if command.stdout:
        devices = command.stdout
        formatted_devices = format_device_string(device_string=devices)
        return formatted_devices


def format_device_string(device_string: str) -> list[dict[str, str | bool]]:
    json_dict = json.loads(device_string)
    # json_dict = remove_duplicate_entries(json_dict=json_dict)
    return json_dict


def remove_duplicate_entries(
    json_list: list[dict[str, str | bool]],
) -> list[dict[str, str | bool]]:
    updated_list = []
    addresses = []
    for device in json_list:
        if device["address"] not in addresses:
            updated_list.append(device)
            addresses.append(device["address"])
        else:
            ...
    return updated_list


def handle_returncodes(errorcode: int) -> int:
    match errorcode:
        case 0:
            return 0
        case 1:
            console.print("1: General failure")
            return 1
        case 64:
            console.print(
                "64: Wrong usage like missing or unexpected arguments, wrong parameters"
            )
            return 1
        case 69:
            console.print("69 Bluetooth or interface not available")
            return 1
        case 70:
            console.print("70 Internal error")
            return 1
        case 71:
            console.print("71 System error like shortage of memory")
            return 1
        case 75:
            console.print("75 Timeout error")
            return 1
        case _:
            console.print("No standart blueutil error")
            return 1


async def device_is_connected(device_address: str) -> bool:
    try:
        command = subprocess.run(
            args=["blueutil", "--is-connected", device_address],
            capture_output=True,
            text=True,
            timeout=TIMEOUT,
        )
    except subprocess.TimeoutExpired as e:
        raise e

    return bool(int(command.stdout))


async def connect_device(device_address: str):
    try:
        command = subprocess.run(
            args=["blueutil", "--connect", device_address],
            capture_output=True,
            text=True,
            timeout=TIMEOUT,
        )
    except subprocess.TimeoutExpired:
        return 1

    returncode = handle_returncodes(errorcode=command.returncode)

    return returncode


async def disconnect_device(device_address: str):
    try:
        command = subprocess.run(
            args=["blueutil", "--disconnect", device_address],
            capture_output=True,
            text=True,
            timeout=TIMEOUT,
        )
    except subprocess.TimeoutExpired:
        return 1

    returncode = handle_returncodes(errorcode=command.returncode)

    return returncode


async def search_new_devices():
    command = subprocess.run(
        args=["blueutil", "--inquiry", "4", "--format", "json"],
        capture_output=True,
        text=True,
        timeout=TIMEOUT,
    )

    handle_returncodes(errorcode=command.returncode)

    if command.stdout:
        devices = command.stdout
        formatted_devices = format_device_string(device_string=devices)
        return formatted_devices


async def pair_device(device_address: str) -> int:
    try:
        command = subprocess.run(
            args=["blueutil", "--pair", device_address],
            capture_output=True,
            text=True,
            timeout=TIMEOUT,
        )
    except subprocess.TimeoutExpired:
        return 1

    returncode = handle_returncodes(errorcode=command.returncode)

    return returncode


async def unpair_device(device_address: str) -> int:
    try:
        command = subprocess.run(
            args=["blueutil", "--unpair", device_address],
            capture_output=True,
            text=True,
            timeout=TIMEOUT,
        )
    except subprocess.TimeoutExpired:
        return 1

    returncode = handle_returncodes(errorcode=command.returncode)

    return returncode



from textual import on, work
from textual.widgets import DataTable, Label
from textual.binding import Binding

from blueutil_tui.utils import (
    get_paired_devices,
    connect_device,
    disconnect_device,
    device_is_connected,
    search_new_devices,
    pair_device,
    unpair_device,
)


class OverViewTable(DataTable):
    BINDINGS = [
        Binding("j, down", "cursor_down", "down", key_display="j/↓"),
        Binding("k, up", "cursor_up", "up", key_display="k/↑"),
        Binding("space", "select_cursor", "dis/connect", key_display="space/enter"),
        Binding("r", "update_devices", "refresh"),
        Binding("s", "display_new_devices", "search"),
        Binding("p", "toggle_pair_device", "un/pair"),
    ]

    def on_mount(self):
        self.show_header = True
        self.cursor_type = "row"
        self.zebra_stripes = True

        self.add_column(":electric_plug: Connection", key="connection")
        self.add_column(":handshake: Paired", key="paired")
        self.add_column(":five_o’clock: Last Access", key="last_access")
        self.add_column(":house: Address", key="address")

        self.action_update_devices()
        return super().on_mount()

    def action_update_devices(self):
        self.clear()
        devices = get_paired_devices()
        for device in devices:
            self.add_row(
                ":green_circle:" if device["connected"] else ":red_circle:",
                ":green_circle:" if device["paired"] else ":red_circle:",
                device["recentAccessDate"],
                device["address"],
                key=device["address"],
                label=f"[blue]{device['name']}[/]",
            )

    @on(DataTable.RowSelected)
    @work(thread=True)
    async def toggle_connection(self, event: DataTable.RowSelected):
        selected_address = event.row_key.value

        if await device_is_connected(device_address=selected_address):
            self.update_cell(
                row_key=selected_address, column_key="connection", value="updating..."
            )

            output = await disconnect_device(device_address=selected_address)
            if output == 0:
                self.update_cell(
                    row_key=selected_address,
                    column_key="connection",
                    value=":red_circle:",
                )
                self.notify(
                    title="Success",
                    message=f"[blue]{self.rows[selected_address].label}[/] disconnected",
                    timeout=1.5,
                )
            else:
                self.update_cell(
                    row_key=selected_address,
                    column_key="connection",
                    value=":green_circle:",
                )
                self.notify(
                    title="Error",
                    message=f"Please check [blue]{self.rows[selected_address].label}[/] if the device is nearby",
                    timeout=1.5,
                    severity="error",
                )
        else:
            self.update_cell(
                row_key=selected_address, column_key="connection", value="updating..."
            )
            output = await connect_device(device_address=selected_address)

            if output == 0:
                self.update_cell(
                    row_key=selected_address,
                    column_key="connection",
                    value=":green_circle:",
                )
                self.notify(
                    title="Success",
                    message=f"[blue]{self.rows[selected_address].label}[/] connected",
                    timeout=1.5,
                )
            else:
                self.update_cell(
                    row_key=selected_address,
                    column_key="connection",
                    value=":red_circle:",
                )
                self.notify(
                    title="Error",
                    message=f"Please check if [blue]{self.rows[selected_address].label}[/] is nearby",
                    timeout=1.5,
                    severity="error",
                )

    @work(thread=True, exclusive=True)
    async def action_display_new_devices(self):
        await self.app.mount(Label("Searching...", id="label-search"))
        new_devices = await search_new_devices()
        await self.app.query_exactly_one("#label-search", Label).remove()

        for device in new_devices:
            if device["address"] in [key.value for key in self.rows.keys()]:
                continue

            self.add_row(
                ":green_circle:" if device["connected"] else ":red_circle:",
                ":green_circle:" if device["paired"] else ":red_circle:",
                device["recentAccessDate"],
                device["address"],
                key=device["address"],
                label=f"[blue]{device['name']}[/]",
            )

    @work(thread=True)
    async def action_toggle_pair_device(self):
        selected_address = self.get_row_at(self.cursor_row)[-1]
        paired = True if "green" in self.get_row_at(self.cursor_row)[1] else False
        # self.notify(f'{selected_address}', timeout=1)
        # return

        if paired:
            self.update_cell(
                row_key=selected_address, column_key="paired", value="updating..."
            )

            output = await unpair_device(device_address=selected_address)
            if output == 0:
                self.update_cell(
                    row_key=selected_address,
                    column_key="paired",
                    value=":red_circle:",
                )
                self.notify(
                    title="Success",
                    message=f"[blue]{self.rows[selected_address].label}[/] unpaired",
                    timeout=1.5,
                )
            else:
                self.update_cell(
                    row_key=selected_address,
                    column_key="paired",
                    value=":green_circle:",
                )
                self.notify(
                    title="Error",
                    message=f"Please check [blue]{self.rows[selected_address].label}[/] if the device is nearby",
                    timeout=1.5,
                    severity="error",
                )
        else:
            self.update_cell(
                row_key=selected_address, column_key="paired", value="updating..."
            )
            output = await pair_device(device_address=selected_address)

            if output == 0:
                self.update_cell(
                    row_key=selected_address,
                    column_key="paired",
                    value=":green_circle:",
                )
                self.notify(
                    title="Success",
                    message=f"[blue]{self.rows[selected_address].label}[/] pairing",
                    timeout=1.5,
                )
            else:
                self.update_cell(
                    row_key=selected_address,
                    column_key="paired",
                    value=":red_circle:",
                )
                self.notify(
                    title="Error",
                    message=f"Please check if [blue]{self.rows[selected_address].label}[/] is nearby",
                    timeout=1.5,
                    severity="error",
                )
